# cython: language_level=3
# distutils: language = c++

from abc import abstractmethod

cimport libc.math as cmath
from libcpp cimport bool
cimport libcpp.complex as ccomplex

from ._implementation cimport RRforce_ecc_corr_NS_v5EHM_v1_flags

cdef:
    double M_EULER_GAMA = 0.577215664901532860606512090082


cdef class BaseForcesCalculation:
    """Base class for the forces calculation"""

    @abstractmethod
    def initialize(self):
        """Updates the internal state given the set of parameters

        .. note::

            What is considered as constant or variable is defined at the time of the
            code generation. The order of elements in the ``variable`` parameter
            is use-case dependent.
        """
        raise NotImplementedError

    @abstractmethod
    def compute(self):
        raise NotImplementedError

    cdef void _initialize(self):
        raise NotImplementedError

    cdef void _compute(self):
        raise NotImplementedError

    cpdef double get(self, str radial_or_azimuthal):
        raise NotImplementedError



cdef class RRforce_ecc_corr_NS_v5EHM_v1_flags(BaseForcesCalculation):
    """
    Eccentricity corrections to the RR force. These equations correspond
    to Eq. (57) of [Gamboa2024]_. See also "docs/mathematica/RR_force/".
    """

    def __cinit__(self):
        self._initialized = False

    cdef void _initialize(self
            , int flagPN1=-1
            , int flagPN2=-1
            , int flagPN3=-1
            , int flagPN32=-1
            , int flagPN52=-1
            , double nu=-1
        ):

        # internal computations intermediate variables declaration/initialisation
        cdef:
            double tmp_init_0 = 74220.126441058865
            double tmp_init_1 = 0.69314718055994531
            double tmp_init_2 = 1.0986122886681097
            double tmp_init_3 = 1.6094379124341004
            double tmp_init_4 = 654562.56399833826
            double tmp_init_5 = 1.9459101490553133
            double tmp_init_6 = cmath.pow(nu, 2)
            double tmp_init_7 = 14844.025288211773
            double tmp_init_8 = 148440.25288211773
            double tmp_init_9 = 1600.0*nu
            double tmp_init_10 = 2375044.0461138837
            double tmp_init_11 = nu*tmp_init_3
            double tmp_init_12 = 7125132.1383416511
            double tmp_init_13 = 20946002.047946824
            double tmp_init_14 = nu*tmp_init_1
            double tmp_init_15 = 1828.0*nu + 4709.0
            double tmp_init_16 = 10825650.0*tmp_init_2
            double tmp_init_17 = 11875.220230569418
            double tmp_init_18 = 5937.6101152847092
            double tmp_init_19 = 23750.440461138837
            double tmp_init_20 = 32728.128199916913
            double tmp_init_21 = 20.0*nu
            double tmp_init_22 = 2968.8050576423546
            double tmp_init_23 = 237504.40461138837
            double tmp_init_24 = 0.33317792380771867
            double tmp_init_25 = cmath.pow(nu, 3)
            double tmp_init_26 = nu*tmp_init_2
            double tmp_init_27 = tmp_init_1*tmp_init_2
            double tmp_init_28 = tmp_init_1*tmp_init_3
            double tmp_init_29 = tmp_init_1*tmp_init_5
            double tmp_init_30 = tmp_init_2*tmp_init_3
            double tmp_init_31 = tmp_init_2*tmp_init_5
            double tmp_init_32 = 5.010191449025495e+18*tmp_init_31
            double tmp_init_33 = cmath.pow(tmp_init_1, 2)
            double tmp_init_34 = cmath.pow(tmp_init_2, 2)
            double tmp_init_35 = cmath.pow(tmp_init_3, 2)
            double tmp_init_36 = cmath.pow(tmp_init_5, 2)
            double tmp_init_37 = 9.8696044010893586
            double tmp_init_38 = 117042170.59249219
            double tmp_init_39 = 129027372.61535244
            double tmp_init_40 = 93633736.473993751
            double tmp_init_41 = 535623421132800.0*nu
            double tmp_init_42 = 5811886080000000.0*tmp_init_11
            double tmp_init_43 = 1032218980.9228195
            double tmp_init_44 = 2340843411.8498438
            double tmp_init_45 = 2809012094.2198125
            double tmp_init_46 = 24773255542.147668
            double tmp_init_47 = 16515503694.765112
            double tmp_init_48 = 1205152697548800.0*tmp_init_26
            double tmp_init_49 = 23625000.0*tmp_init_3
            double tmp_init_50 = 1.401685291302912e+17*tmp_init_27
            double tmp_init_51 = 93884313600.0*tmp_init_37
            double tmp_init_52 = 11340.0*tmp_init_2
            double tmp_init_53 = 1597440.0*nu
            double tmp_init_54 = -1008000000.0*tmp_init_3
            double tmp_init_55 = 29260542.648123047
            double tmp_init_56 = 516109490.46140975
            double tmp_init_57 = 880165440.0*tmp_init_37
            double tmp_init_58 = 3096656942.7684585
            double tmp_init_59 = 16854072565.318875
            double tmp_init_60 = 8257751847.3825561
            double tmp_init_61 = -1.2917493386243386e-7*flagPN32

        # computations
        self._gr_k_0 = 625444092000.0*nu - 4217291001000.0
        self._gr_k_1 = 2914507008000.0*nu + 33690890520000.0
        self._gr_k_2 = -5872947444000.0*nu - 51825960837000.0
        self._gr_k_3 = 67011806736000.0*nu + 193697612388000.0
        self._gr_k_4 = -17318643048000.0*nu - 67566853242000.0
        self._gr_k_5 = -952383096000.0*nu - 704469150000.0
        self._gr_k_6 = 1189448064000.0*nu + 7139382768000.0
        self._gr_k_7 = -4225229568000.0*nu - 14944142304000.0
        self._gr_k_8 = 33638758272000.0*nu + 77884843680000.0
        self._gr_k_9 = -7628535936000.0*nu - 23632578144000.0
        self._gr_k_10 = 4776081408000.0*nu + 12303373824000.0
        self._gr_k_11 = -2556997632000.0*nu - 3086512128000.0
        self._gr_k_12 = tmp_init_0*(3394527768.0*nu - 2931098059.0)
        self._gr_k_13 = 13196445625394.797 - 18621224087829.91*nu
        self._gr_k_14 = -94500.0*nu*(853502374336.0*tmp_init_1 + 414870587910.0*tmp_init_2 - 453591625000.0*tmp_init_3 - 163172692305.0*tmp_init_5) + 12737879219016000.0*tmp_init_1 + 10718635338195750.0*tmp_init_2 - 8193414576375000.0*tmp_init_3 - tmp_init_4*(3452644.0*nu + 30467937.0) + 34484805791100.0
        self._gr_k_15 = -1375102315322800.0*nu - 7709909711411250.0*tmp_init_5 - 126162787915200.0*tmp_init_6
        self._gr_k_16 = -tmp_init_7*(276850224.0*nu - 492515813.0)
        self._gr_k_17 = -4.0*nu*(40496086870948800.0*tmp_init_1 + 11302168846364550.0*tmp_init_2 - 7770376195312500.0*tmp_init_3 - 14379165472903875.0*tmp_init_5 - 3584579256938.0) + 53065019667643200.0*tmp_init_1 + 19125904617933750.0*tmp_init_2 - 12710014376062500.0*tmp_init_3 + tmp_init_4*(367692.0*nu - 2479301.0) - 19194680315823075.0*tmp_init_5 + 953459738400.0*tmp_init_6 + 4120253808300.0
        self._gr_k_18 = 8.0*nu*(26603407189800000.0*tmp_init_1 + 12809628368826975.0*tmp_init_2 - 9271198543359375.0*tmp_init_3 - 9040258747102275.0*tmp_init_5 + 8343043960136.0) + 1121532733655.2648*nu - 57155843890857600.0*tmp_init_1 - 33048921054041550.0*tmp_init_2 + 22702518658968750.0*tmp_init_3 + 20272625125568550.0*tmp_init_5 - 31064454811200.0*tmp_init_6 - 10234360268915.633
        self._gr_k_19 = -tmp_init_8*(1484090728.0*nu - 377673245.0)
        self._gr_k_20 = tmp_init_8*(408785852.0*nu - 422250829.0)
        self._gr_k_21 = tmp_init_7*(458452124.0*nu - 714847261.0)
        self._gr_k_22 = 25786843054248.839*nu - 215504510256000.0*tmp_init_1 - 1648108030968000.0*tmp_init_2 + 1508705068500000.0*tmp_init_3 + 15703674076800.0*tmp_init_6 + tmp_init_9*(9494459762760.0*tmp_init_1 + 1831862692800.0*tmp_init_2 - 5252410406250.0*tmp_init_3 - 1236647171603.0) - 41728615795880.794
        self._gr_k_23 = -300.0*nu*(-139722832476864.0*tmp_init_1 - 66327553134108.0*tmp_init_2 + 53484168937500.0*tmp_init_3 + 42983578393812.0*tmp_init_5 + 65794866688.0) - 366486961332.4136*nu - 13224505085668800.0*tmp_init_1 - 6800170185062100.0*tmp_init_2 + 5864039224312500.0*tmp_init_3 + 3701782451511900.0*tmp_init_5 + 2212289760000.0*tmp_init_6 - 1208503990079.9118
        self._gr_k_24 = -tmp_init_0*(682334308.0*nu + 80530829.0)
        self._gr_k_25 = -2000.0*nu*(26134293901968.0*tmp_init_1 + 10766380561335.0*tmp_init_2 - 13885498659375.0*tmp_init_3 - 3887349434325.0*tmp_init_5 + 32673748.0) - 6664394708095.7531*nu + 11767784140728000.0*tmp_init_1 + 6907439499943500.0*tmp_init_2 - 7543757422687500.0*tmp_init_3 - 1943674717162500.0*tmp_init_5 + 32447363222400.0*tmp_init_6 + 22565928285143.119
        self._gr_k_26 = 16000.0*nu*(948435578496.0*tmp_init_1 + 43437465729.0*tmp_init_2 - 13383466550.0) - 813668512356000.0*tmp_init_2 + 1737314743500000.0*tmp_init_3 + 8963382153600.0
        self._gr_k_27 = tmp_init_10*(53167044.0*nu - 56685535.0)
        self._gr_k_28 = -14400.0*nu*(860896552320.0*tmp_init_1 + 309918083580.0*tmp_init_2 - 455275406250.0*tmp_init_3 - 105063498225.0*tmp_init_5 - 1027861228.0) + 457712036751.73401*nu + 3802279416768000.0*tmp_init_1 + 1351566822312000.0*tmp_init_2 - 2097343678500000.0*tmp_init_3 - 378228593610000.0*tmp_init_5 - 4955548300800.0*tmp_init_6 - 1238707302788.2226
        self._gr_k_29 = -tmp_init_10*(4447812.0*nu - 2799827.0)
        self._gr_k_30 = -3251824925939.6486*nu - 5541534625536000.0*tmp_init_1 - 14012667900000000.0*tmp_init_11 - 35738928153600.0*tmp_init_6 - 11501324048515.314
        self._gr_k_31 = -tmp_init_12*(15742124.0*nu - 6174963.0)
        self._gr_k_32 = tmp_init_12*(3455952.0*nu - 5858833.0)
        self._gr_k_33 = -44800.0*nu*(76522767120.0*tmp_init_1 - 22048382655.0*tmp_init_2 - 19307109375.0*tmp_init_3 + 10761018208.0) + 574601392512000.0*tmp_init_1 + tmp_init_13*(617996.0*nu + 1430865.0) + 102579485400000.0*tmp_init_2 - 216239625000000.0*tmp_init_3 + 2758456166400.0*tmp_init_6 - 41383276099200.0
        self._gr_k_34 = 6400.0*nu*(638590554000.0*tmp_init_1 + 22647598785.0*tmp_init_2 - 292864359375.0*tmp_init_3 - 1132554796.0) - 1208678599296000.0*tmp_init_1 - tmp_init_13*(140148.0*nu + 434167.0) - 41897303784000.0*tmp_init_2 + 516564783000000.0*tmp_init_3 + 5354443776000.0*tmp_init_6 + 15691820860800.0
        self._gr_k_35 = -13837006612659.486*nu - 2991949861871.7344
        self._gr_k_36 = 33576130682378.556*nu - 54988919790865.526
        self._gr_k_37 = 384000.0*nu*(861942627.0*tmp_init_2 + 717609375.0*tmp_init_3 - 115128058.0) - 983959392204.35003*nu + 294102918144000.0*tmp_init_1 - 2472633139200.0*tmp_init_6 + 877277989473.22327
        self._gr_k_38 = -2333510000640000.0*tmp_init_14 - 176450799168000.0*tmp_init_2 - 137781000000000.0*tmp_init_3
        self._gr_k_39 = 22043334802201.653 - 32658185658731.724*nu
        self._gr_k_40 = 12902400.0*nu*(18920520.0*tmp_init_1 - tmp_init_16 - 4294151.0) - 23594747904000.0*tmp_init_1 + 1005408098.3014476*tmp_init_15 + 33837543936000.0*tmp_init_2 - 7947878400.0*tmp_init_6 - 3867691161600.0
        self._gr_k_41 = 4227882407720.6155*nu - 21209466337787.253
        self._gr_k_42 = 3.4173262926569805e-12*flagPN52
        self._gr_k_43 = 875221653600.0*nu - 795273633000.0
        self._gr_k_44 = -264215989800.0*nu - 801448007850.0
        self._gr_k_45 = 377622255150.0 - 12441102093000.0*nu
        self._gr_k_46 = 5165017099200.0*nu + 30924847472400.0
        self._gr_k_47 = 759076353000.0 - 52201648800.0*nu
        self._gr_k_48 = -8859138674400.0*nu - 13737381121800.0
        self._gr_k_49 = 327718742400.0*nu - 419903416800.0
        self._gr_k_50 = -6156182995200.0*nu - 231159096000.0
        self._gr_k_51 = 1290060172800.0*nu + 14947413667200.0
        self._gr_k_52 = -3609690739200.0*nu - 6004029225600.0
        self._gr_k_53 = 955216281600.0*nu + 2460674764800.0
        self._gr_k_54 = -2202884812800.0*nu - 846178099200.0
        self._gr_k_55 = tmp_init_17*(5154379.0*nu - 21959928.0)
        self._gr_k_56 = tmp_init_18*(1290809182.0*nu - 13452056029.0)
        self._gr_k_57 = -tmp_init_19*(4744123.0*nu - 121928598.0)
        self._gr_k_58 = 4629506863627800.0*tmp_init_1 + 4794307052782275.0*tmp_init_2 - tmp_init_20*(146279860.0*nu - 4440003.0) - tmp_init_21*(1883219514903720.0*tmp_init_1 + 892452729397635.0*tmp_init_2 - 1005084201515625.0*tmp_init_3 - 342682110043875.0*tmp_init_5 - 5265926028868.0) - 3197253222253125.0*tmp_init_3 - 1713410550219375.0*tmp_init_5 + 4799617305000.0*tmp_init_6 + 12369794712975.0
        self._gr_k_59 = 56.0*nu*(771055284998160.0*tmp_init_1 + 371494865148255.0*tmp_init_2 - 222184605234375.0*tmp_init_3 - 300646037636595.0*tmp_init_5 - 57975725792.0) + 336794432479.29538*nu - 12419366589655200.0*tmp_init_1 - 6643053728689770.0*tmp_init_2 + 4376677012706250.0*tmp_init_3 + 4548237047194050.0*tmp_init_5 + 3500617382400.0*tmp_init_6 - 297483280053.83508
        self._gr_k_60 = -28.0*nu*(619971514657560.0*tmp_init_1 + 66541031914365.0*tmp_init_2 - 28617799921875.0*tmp_init_3 - 234722194617015.0*tmp_init_5 + 62094118466.0) + 8090768243912760.0*tmp_init_1 + 2285751830903415.0*tmp_init_2 - tmp_init_20*(3106596.0*nu + 9423257.0) - 1774631927053125.0*tmp_init_3 - 2705387336894025.0*tmp_init_5 + 191268302280.0*tmp_init_6 + 634158733635.0
        self._gr_k_61 = -tmp_init_17*(1351060342.0*nu + 3156310921.0)
        self._gr_k_62 = tmp_init_22*(1501193028.0*nu - 4773871337.0)
        self._gr_k_63 = 80.0*nu*(169754320708560.0*tmp_init_1 + 1745635439205.0*tmp_init_2 - 74190600421875.0*tmp_init_3 + 4550889179476.0) + 1987552519428.343*nu - 515738992104000.0*tmp_init_1 - 1090769846084100.0*tmp_init_2 + 1015977681337500.0*tmp_init_3 - 2818744968960.0*tmp_init_6 - 14188544646440.097
        self._gr_k_64 = -tmp_init_22*(30386972.0*nu + 117233297.0)
        self._gr_k_65 = -3409089080265.787*nu + 2619164066752800.0*tmp_init_1 + 1530781594453125.0*tmp_init_2 - tmp_init_21*(543750482247840.0*tmp_init_1 + 279982034152065.0*tmp_init_2 - 292874131265625.0*tmp_init_3 - 108950847659325.0*tmp_init_5 + 216300611096.0) - 1541622365128125.0*tmp_init_3 - 544754238296625.0*tmp_init_5 - 2347612166400.0*tmp_init_6 + 11036512125442.965
        self._gr_k_66 = -tmp_init_18*(1831769596.0*nu + 12488714761.0)
        self._gr_k_67 = 60.0*nu*(65519029394784.0*tmp_init_1 + 37124246571993.0*tmp_init_2 - 14158020515625.0*tmp_init_3 - 32582292069537.0*tmp_init_5 + 31786967336.0) - 20087739614.032203*nu - 2207870474009760.0*tmp_init_1 - 899614976769495.0*tmp_init_2 + 860169263034375.0*tmp_init_3 + 584337195438255.0*tmp_init_5 - 368367467520.0*tmp_init_6 - 185674274228.04955
        self._gr_k_68 = 4092595338769.4002*nu - 51883218524429.833
        self._gr_k_69 = -480.0*nu*(1945830282480.0*tmp_init_1 + 1683606560580.0*tmp_init_2 - 1224023062500.0*tmp_init_3 - 630380989350.0*tmp_init_5 + 2570900939.0) + 126109594530.07345*nu + 669960533289600.0*tmp_init_1 + 159426150156000.0*tmp_init_2 - 308740779900000.0*tmp_init_3 - 75645718722000.0*tmp_init_5 + 340467054720.0*tmp_init_6 + 94197510851.575011
        self._gr_k_70 = -tmp_init_19*(4067052.0*nu - 75431789.0)
        self._gr_k_71 = 640.0*nu*(5211509761080.0*tmp_init_1 + 1011939367005.0*tmp_init_2 - 2947044515625.0*tmp_init_3 + 58870094156.0) - 2368963507219.9187*nu - 605722589740800.0*tmp_init_1 - 206667397188000.0*tmp_init_2 + 398485804500000.0*tmp_init_3 - 32705393280.0*tmp_init_6 + 7189899239882.8818
        self._gr_k_72 = -tmp_init_19*(398233636.0*nu + 852822433.0)
        self._gr_k_73 = 496428626937.15892*nu + 138270888000000.0*tmp_init_1 + 48365460385200.0*tmp_init_2 - 71665256250000.0*tmp_init_3 - 506843481600.0*tmp_init_6 - tmp_init_9*(835931692512.0*tmp_init_1 - 268111553661.0*tmp_init_2 - 179163140625.0*tmp_init_3 - 55848045928.0) - 5599690010939.2616
        self._gr_k_74 = tmp_init_23*(9127700.0*nu - 22160837.0)
        self._gr_k_75 = -tmp_init_23*(21309308.0*nu + 178176097.0)
        self._gr_k_76 = 11200.0*nu*(32273943264.0*tmp_init_1 + 13149190971.0*tmp_init_2 - 23321165625.0*tmp_init_3 - 21801176.0) - 1389046693410.8224*nu - 223273365235200.0*tmp_init_1 + 15068164935600.0*tmp_init_2 + 74895625350000.0*tmp_init_3 + 5995737600.0*tmp_init_6 + 4566685701904.707
        self._gr_k_77 = 1407470102079.4564*nu - 24942663171371.847
        self._gr_k_78 = -15360.0*nu*(8318313360.0*tmp_init_1 + 156534525.0*tmp_init_2 - 3588046875.0*tmp_init_3 - 647999071.0) - 847693081281.22716*nu + 56159773286400.0*tmp_init_1 - 19599145171200.0*tmp_init_2 - 13778100000000.0*tmp_init_3 - 497805127680.0*tmp_init_6 + 3501308988323.4378
        self._gr_k_79 = -4605959019298.1556*nu - 8111782436154.4651
        self._gr_k_80 = 2580480.0*nu*(18920520.0*tmp_init_1 - tmp_init_16 + 3376819.0) - 4718949580800.0*tmp_init_1 + 201081619.66028951*tmp_init_15 + 6767508787200.0*tmp_init_2 - 1589575680.0*tmp_init_6 - 773538232320.0
        self._gr_k_81 = -568414541468.34356*nu - 17947204637877.541
        self._gr_k_82 = 1.7086631463284902e-11*flagPN52
        self._gr_k_83 = 721234922781643.51*nu - 1.3636557722672427e+20*tmp_init_1 - 6.3527831965789059e+19*tmp_init_2 + 4.8032921759181545e+19*tmp_init_3 + 4.4718062307761151e+19*tmp_init_5 - 667140113275378.25
        self._gr_k_84 = 7.6484890055839854e+17*nu + 1.0003102839269806e+20*tmp_init_1 - 27001054080000000.0*tmp_init_11 + 1.248134545760256e+17*tmp_init_14 - 2.3342199781535711e+18*tmp_init_2 - 13596795717120000.0*tmp_init_24 + 25422889089484800.0*tmp_init_25 - 44863209383731200.0*tmp_init_26 - 8.3870706332861522e+20*tmp_init_27 - 1.0402029880704e+20*tmp_init_28 + 3.6493987097840026e+19*tmp_init_29 - 3.7208248806500674e+19*tmp_init_3 + 1.82986394976e+20*tmp_init_30 + tmp_init_32 + 3.6513382891019747e+20*tmp_init_33 + 1.0464461315660943e+19*tmp_init_34 + 4.45121667144e+19*tmp_init_35 + 1.7914527141037148e+19*tmp_init_36 - 146694240.0*tmp_init_37*(77320875.0*nu - 51197152.0) - 3.6949385251453989e+18*tmp_init_5 + 1.823731993209168e+17*tmp_init_6 + 3.9865141605688158e+17
        self._gr_k_85 = -tmp_init_38*(-14563300.0*nu + 641286641520.0*tmp_init_1 + 65747154060.0*tmp_init_2 - 320642437500.0*tmp_init_3 + 5616819193.8383162)
        self._gr_k_86 = 31117806720000000.0*nu*tmp_init_3 + 30727275018570600.0*nu + 2.4909881635802161e+20*tmp_init_1*tmp_init_2 + 1.3175724570624e+20*tmp_init_1*tmp_init_3 - 5.1277787319896654e+19*tmp_init_1 - 70594190508441600.0*tmp_init_14 - 2.5747687580603079e+19*tmp_init_2 - 681740943206400.0*tmp_init_24 - 3110779414195200.0*tmp_init_25 - 1267921067212800.0*tmp_init_26 - 3.4019818481037312e+19*tmp_init_29 + 1.561659552684e+19*tmp_init_3 - 8.884742017536e+19*tmp_init_30 - 1.002038289805099e+19*tmp_init_31 - 1.4064916041022759e+20*tmp_init_33 + 4.759024646478187e+19*tmp_init_34 - 1.62449201376e+19*tmp_init_35 - 1.5857874478319892e+19*tmp_init_36 - 73347120.0*tmp_init_37*(10068165.0*nu - 9789472.0) - tmp_init_39*(808600.0*nu - 273209546904.0*tmp_init_1 - 141823395882.0*tmp_init_2 + 76881000000.0*tmp_init_3 + 113797171740.0*tmp_init_5 + 1160229.0) + 1.9876594399341217e+19*tmp_init_5 + 16472333722413600.0*tmp_init_6 + 31736149385163111.0
        self._gr_k_87 = -tmp_init_40*(-25125800.0*nu - 2228211294120.0*tmp_init_1 - 955325511315.0*tmp_init_2 + 1187722265625.0*tmp_init_3 + 350471076795.0*tmp_init_5 + 468057993.53571504)
        self._gr_k_88 = 76987908974592000.0*nu*tmp_init_1 + 24997986052300800.0*nu*tmp_init_2 + 2.386443017064519e+18*nu + 5.9139645428141752e+20*tmp_init_1*tmp_init_2 - 3.3318983738477961e+20*tmp_init_1 - 50611841280000000.0*tmp_init_11 + 4.5144542177077078e+19*tmp_init_2 + 93683541995366400.0*tmp_init_24 - 5815685250892800.0*tmp_init_25 - 2.515278116736e+19*tmp_init_28 + 1.1604456894492e+20*tmp_init_3 - 7.785239349888e+19*tmp_init_30 - 1.1600118175286108e+20*tmp_init_33 - 1.0655604626248835e+20*tmp_init_34 - 3.77803680408e+19*tmp_init_35 - 1100206800.0*tmp_init_37*(147341413.0*nu + 34914272.0) - tmp_init_39*(4701800.0*nu - 569935528680.0*tmp_init_1 - 49700651490.0*tmp_init_2 + 276705187500.0*tmp_init_3 - 3168912213.0) - 5.804234988029896e+17*tmp_init_6 + 1.9033377790404216e+18
        self._gr_k_89 = -2.618667297523985e+18*nu + 1.7181344309512374e+20*tmp_init_1 + 42378336000000000.0*tmp_init_11 - 1.619059487367168e+17*tmp_init_14 - 3.4271471272343534e+19*tmp_init_2 + 1.190724422114304e+17*tmp_init_24 + 49817921698252800.0*tmp_init_25 + 50044738299494400.0*tmp_init_26 - 8.1744412408101274e+19*tmp_init_27 + 8.31815018496e+18*tmp_init_28 - 5.1961937540377482e+19*tmp_init_3 + 31685955840000000.0*tmp_init_30 - 2.4459733548498125e+19*tmp_init_33 + 3.7288946655478426e+19*tmp_init_34 + 8.6608279296e+18*tmp_init_35 + 733471200.0*tmp_init_37*(128839671.0*nu + 113328032.0) + 4.5097680817176288e+17*tmp_init_6 + 6.3474047207131503e+18
        self._gr_k_90 = -1576518806016000.0*nu + 1.9110002109419612e+20*tmp_init_1 + 1.6664684109326669e+19*tmp_init_2 - 9.277955927856e+19*tmp_init_3 + 1.0035125409412536e+18
        self._gr_k_91 = -271124485632000.0*nu + 9.1607467071603732e+19*tmp_init_1 + 4.7553543481437988e+19*tmp_init_2 - 2.577828540672e+19*tmp_init_3 - 3.8156319137254349e+19*tmp_init_5 - 1123965668214121.0
        self._gr_k_92 = -3502233610444800.0*nu - 2.1970982913991373e+20*tmp_init_1 - 1.0123141365395638e+20*tmp_init_2 + 1.1623684842e+20*tmp_init_3 + 3.9223304353250519e+19*tmp_init_5 + 75350367393665595.0
        self._gr_k_93 = 3927556375142400.0*nu - 4.5063047054037197e+19*tmp_init_1 + 1.0922981684967475e+19*tmp_init_2 + 1.304229148992e+19*tmp_init_3 - 1.8971577181161268e+17
        self._gr_k_94 = 1390238915788800.0*nu + 7.0881946251514675e+19*tmp_init_1 + 7.3946199560454144e+18*tmp_init_2 - 3.547770855552e+19*tmp_init_3 + 2.8295854200732395e+17
        self._gr_k_95 = 629078549299200.0*nu - 2.491775455652905e+19*tmp_init_1 - 1.4993561094853018e+19*tmp_init_2 + 1.423755615744e+19*tmp_init_3 + 5.5668793878061056e+18*tmp_init_5 - 947087432559709.94
        self._gr_k_96 = 2733482026598400.0*nu - 3.656651911422935e+19*tmp_init_1 + 8.4038151146729472e+18*tmp_init_2 + 1.122739035264e+19*tmp_init_3 + 3.6957037176829488e+17
        self._gr_k_97 = -285837498777600.0*nu - 1.9720933334581248e+19*tmp_init_1 + 5.3050503075913728e+18*tmp_init_2 + 5.21762072832e+18*tmp_init_3 + 5.410307487474739e+17
        self._gr_k_98 = -590022674841600.0*nu + 9.415274762944512e+18*tmp_init_1 + 1.9787854073315328e+18*tmp_init_2 - 5.42886043392e+18*tmp_init_3 + 21852145899610126.0
        self._gr_k_99 = 1069851989606400.0*nu + 6.370989520896e+18*tmp_init_1 - 3.2864774792306688e+18*tmp_init_2 - 1.024097055262311e+17
        self._gr_k_100 = 1484123229388800.0*nu + 3.4718749642063872e+18*tmp_init_1 - 1.5615650198421504e+18*tmp_init_2 + 1.9421359821689416e+17
        self._gr_k_101 = -256652889292800.0*nu - 3.2704908583108608e+18*tmp_init_1 + 6.16887537057792e+17*tmp_init_2 + 1.01395058688e+18*tmp_init_3 + 1.0370286555337401e+17
        self._gr_k_102 = 8.260858221428736e+17*tmp_init_1 - 4.102849654751232e+17*tmp_init_2 + tmp_init_41 + 1.8642820757651128e+17
        self._gr_k_103 = -58695793463762100.0*nu + 6.4161977038087864e+18*tmp_init_1 + 26741185280409600.0*tmp_init_14 + 9.6539442666081085e+18*tmp_init_2 + 1836377174016000.0*tmp_init_24 + 7079205359116800.0*tmp_init_25 - 7431774968217600.0*tmp_init_26 - 1.0051901349254038e+20*tmp_init_27 - 1.246783684608e+19*tmp_init_28 + 5.9380046803265126e+18*tmp_init_29 - 6.16350518784e+18*tmp_init_3 + 2.237028482304e+19*tmp_init_30 + 4.3910720940648038e+19*tmp_init_33 - 1.5016466238610637e+18*tmp_init_34 + 5.3132653728e+18*tmp_init_35 + 2.5978770476428493e+18*tmp_init_36 + 97796160.0*tmp_init_37*(19824525.0*nu - 8753152.0) - tmp_init_42 + tmp_init_43*(234520.0*nu - 9289319760.0*tmp_init_1 - 5589588060.0*tmp_init_2 + 5307750000.0*tmp_init_3 + 2075328360.0*tmp_init_5 - 1764543.0) - 2.6102478907268628e+18*tmp_init_5 - 24353766821210400.0*tmp_init_6 - 43721523806087583.0
        self._gr_k_104 = tmp_init_44*(1565000.0*nu - 25843717704.0*tmp_init_1 - 1731233574.0*tmp_init_2 + 12426225000.0*tmp_init_3 - 150711551.18760273)
        self._gr_k_105 = 15833079830937600.0*nu*tmp_init_2 + 46392784668762979.0*nu + 1.9166269917139108e+20*tmp_init_1*tmp_init_2 + 1.1951489846524914e+19*tmp_init_1 - 24235314595430400.0*tmp_init_14 - 5.152834470463527e+18*tmp_init_2 + 3337024042598400.0*tmp_init_24 - 5621251934246400.0*tmp_init_25 - 1.449104380416e+19*tmp_init_28 - 1.1605495297715185e+18*tmp_init_3 - 2.379615283584e+19*tmp_init_30 - 1.7512999601061888e+19*tmp_init_33 - 3.2921398862831002e+19*tmp_init_34 - 1.51212422592e+19*tmp_init_35 + 488980800.0*tmp_init_37*(11033469.0*nu - 603776.0) - 2.530034186744232e+17*tmp_init_6 + 3.1925711414239145e+18
        self._gr_k_106 = tmp_init_40*(-3081800.0*nu + 454960471560.0*tmp_init_1 + 171501266790.0*tmp_init_2 - 237798750000.0*tmp_init_3 - 62259850800.0*tmp_init_5 - 48571644.065082784)
        self._gr_k_107 = 4.0155005116042154e+18*nu - 1.2771679486387329e+19*tmp_init_1 + 11623772160000000.0*tmp_init_11 + 5168537115033600.0*tmp_init_14 + 5.355541284265769e+18*tmp_init_2 + 1.38438053462016e+17*tmp_init_24 + 61323106876646400.0*tmp_init_25 - 14894451287654400.0*tmp_init_26 - 7.5972924551533363e+19*tmp_init_27 + 1.12895887104e+19*tmp_init_28 + 2.9043286615387687e+18*tmp_init_3 + 2.50319051136e+18*tmp_init_30 - 1.3408987282971034e+19*tmp_init_33 + 3.0162469734448128e+19*tmp_init_34 + 5.3484719904e+18*tmp_init_35 - 977961600.0*tmp_init_37*(183175659.0*nu - 42965312.0) + 5.1744795712386784e+17*tmp_init_6 + 2.2505441126258111e+18
        self._gr_k_108 = -tmp_init_45*(134280.0*nu + 2033966200.0*tmp_init_1 - 1061475030.0*tmp_init_2 + 102812500.0*tmp_init_3 + 14748756.82610618)
        self._gr_k_109 = 8.016126737390976e+17*nu - 5.9831907591165051e+19*tmp_init_1 - 1263521916518400.0*tmp_init_14 + 1.0340079258432946e+19*tmp_init_2 + 35618957505331200.0*tmp_init_24 + 14983256413900800.0*tmp_init_25 + 778327783833600.0*tmp_init_26 - 2.6320534914465792e+19*tmp_init_27 + 3.06062862336e+18*tmp_init_28 + 2.030589171072e+19*tmp_init_3 + 1.52092588032e+18*tmp_init_30 - 1.5406136462042726e+19*tmp_init_33 + 1.2500368136279654e+19*tmp_init_34 + 2.757851712e+18*tmp_init_35 - 46942156800.0*tmp_init_37*(1639221.0*nu + 294368.0) - tmp_init_46*(4440.0*nu + 306331200.0*tmp_init_1 - 82404945.0*tmp_init_2 - 81046875.0*tmp_init_3 - 9552524.0) - 2.96530086001536e+17*tmp_init_6 + 1.4113236471649056e+18
        self._gr_k_110 = -700642523287600.44*nu - 1.5925494066518119e+19*tmp_init_1 - 3.860091610449495e+17*tmp_init_2 + 7.1303260746652166e+18*tmp_init_3 - 13848212206231937.0
        self._gr_k_111 = 4042283006361600.0*nu*tmp_init_2 + 1.7276135882364582e+17*nu + 2.6802364233351168e+19*tmp_init_1*tmp_init_2 + 1.8283494385490419e+19*tmp_init_1 - 8173836582912000.0*tmp_init_14 - 4.43529796749274e+18*tmp_init_2 - 4637134017331200.0*tmp_init_24 + 8809613717760000.0*tmp_init_25 - 3.02307489792e+18*tmp_init_28 - 4.9235781496951713e+18*tmp_init_3 - 3.04185176064e+18*tmp_init_30 - 2.190313525542912e+18*tmp_init_33 - 3.840033662631936e+18*tmp_init_34 - 2.3119012224e+18*tmp_init_35 - 11735539200.0*tmp_init_37*(45633.0*nu - 257120.0) + 56765001757708800.0*tmp_init_6 + 1.4415247695838723e+17
        self._gr_k_112 = -3987748475943625.0*nu + 1.4025096260343024e+19*tmp_init_1 - 4.5066681041687874e+18*tmp_init_2 - 2.9568363556781301e+18*tmp_init_3 - 3.3859369067664321e+17
        self._gr_k_113 = -1.1655277844193271e+18*nu + 4.7909654354313111e+19*tmp_init_1 + 13440800224051200.0*tmp_init_14 - 6.4441847811322757e+18*tmp_init_2 + 62391759446016000.0*tmp_init_24 + 23231913477580800.0*tmp_init_25 - 5674260617625600.0*tmp_init_26 + 5.8471691283726336e+18*tmp_init_27 - 1.68174468e+19*tmp_init_3 + 6.6139861588770816e+18*tmp_init_33 - 5.38864039397376e+18*tmp_init_34 + 35206617600.0*tmp_init_37*(1189615.0*nu + 1206112.0) + 1.5404564350586112e+17*tmp_init_6 + 3.1801496969550637e+18
        self._gr_k_114 = 1848105237029099.1*nu + 5.1941878592895576e+18*tmp_init_1 - 1.5720683376704364e+18*tmp_init_2 - 1.0618065716150891e+18*tmp_init_3 - 1.2811940432486156e+17
        self._gr_k_115 = 1.5872483442053136e+18*nu - 2.2153159232638157e+18*tmp_init_1 + 424035208396800.0*tmp_init_14 + 1.8924463560580301e+18*tmp_init_2 + 56691103924224000.0*tmp_init_24 + 36045502132147200.0*tmp_init_25 + 2410305395097600.0*tmp_init_26 + 2.2981798421987328e+18*tmp_init_27 + 4.0095842265464832e+18*tmp_init_33 - 2.283213931536384e+18*tmp_init_34 - 844958822400.0*tmp_init_37*(95735.0*nu - 25616.0) + tmp_init_47*(34580.0*nu + 80894520.0*tmp_init_1 - 36384390.0*tmp_init_2 + 5173701.0) + 2.024195211107712e+17*tmp_init_6 + 9.2250323034864624e+17
        self._gr_k_116 = 959658629529600.0*nu*tmp_init_1 + 97505971495300800.0*nu + 9.0128941056e+17*tmp_init_1*tmp_init_3 + 1.5744881652085555e+18*tmp_init_1 - 1.8748554721763328e+17*tmp_init_2 - 630902587392000.0*tmp_init_24 - 764216698291200.0*tmp_init_25 - 4.0678075224686592e+18*tmp_init_27 - 3.8358448128e+17*tmp_init_3 - 3.1381455212642304e+18*tmp_init_33 + 1.8619782997229568e+18*tmp_init_34 + 5.633058816e+17*tmp_init_35 - 281652940800.0*tmp_init_37*(2255.0*nu + 2512.0) - tmp_init_47*(5980.0*nu + 76202280.0*tmp_init_1 - 14373450.0*tmp_init_2 - tmp_init_49 - 2934141.0) - tmp_init_48 - 78879289204569600.0*tmp_init_6 + 9.8906544059520758e+17
        self._gr_k_117 = 1578440075983997.0*nu - 4.6198496059924785e+17*tmp_init_1 - 1.0014959583473521e+17*tmp_init_2 - 56244853902306136.0
        self._gr_k_118 = -15056025792007731.0*nu + 5.6281774493268373e+18*tmp_init_1 + 1071246842265600.0*tmp_init_14 - 2.8734061607904327e+18*tmp_init_2 + 9409461446246400.0*tmp_init_24 + 14771618801049600.0*tmp_init_25 + 1.284878183694336e+18*tmp_init_33 - 4.628481638989824e+17*tmp_init_34 + tmp_init_50 - tmp_init_51*(229887.0*nu + 9376.0) - 1.055072151834624e+17*tmp_init_6 + 1.0151085202106022e+18
        self._gr_k_119 = -690342812275461.12*nu - 1.10563794689136e+18*tmp_init_1 + 7.2933369791097242e+17*tmp_init_2 - 1.3216808590728387e+17
        self._gr_k_120 = -2.959795247746176e+17*nu - 6.1608774899892714e+18*tmp_init_1 + 3.7396800212824347e+18*tmp_init_2 + 18927077621760000.0*tmp_init_24 + 11845894756147200.0*tmp_init_25 - 2.803370582605824e+17*tmp_init_27 + 35474751199641600.0*tmp_init_33 + 3.606419447414784e+17*tmp_init_34 + tmp_init_51*(120909.0*nu + 171616.0) - 6106585598668800.0*tmp_init_6 + 1.0207007578270931e+18
        self._gr_k_121 = 2.049647046792576e+17*nu - 2.1872182519151081e+17*tmp_init_1 + 3.6678822349897728e+17*tmp_init_2 + 5191427004825600.0*tmp_init_24 + 17659059283353600.0*tmp_init_25 - 41531416038604800.0*tmp_init_33 + 70084264565145600.0*tmp_init_34 - 3379835289600.0*tmp_init_37*(6847.0*nu - 2304.0) + 2378232532046.1761*tmp_init_52 + 15630815986329600.0*tmp_init_6 + 2.587183173231439e+17
        self._gr_k_122 = 1.1557515870728424e-15*flagPN3
        self._gr_k_123 = -57647649504.0*nu - 10477437792.0*tmp_init_6 + 25551077982.0
        self._gr_k_124 = -5892492624.0*nu - 226723824.0*tmp_init_6 + 10425470991.0
        self._gr_k_125 = -122817381936.0*nu + 23255029488.0*tmp_init_6 + 83299288385.0
        self._gr_k_126 = 222931895904.0*nu - 3458116512.0*tmp_init_6 - 185351458286.0
        self._gr_k_127 = -155267082960.0*nu - 6081208032.0*tmp_init_6 + 122009866050.0
        self._gr_k_128 = 50545157880.0*nu + 7247672784.0*tmp_init_6 - 37830602731.0
        self._gr_k_129 = 10959425112.0*nu + 2308597776.0*tmp_init_6 - 12418573799.0
        self._gr_k_130 = -83631629952.0*nu + 10669116672.0*tmp_init_6 + 67980197712.0
        self._gr_k_131 = -17924806272.0*nu - 6981102336.0*tmp_init_6 + 12555666256.0
        self._gr_k_132 = 143311401984.0*nu + 1998188544.0*tmp_init_6 - 135413136000.0
        self._gr_k_133 = 15726595200.0*nu + 7947078912.0*tmp_init_6 - 23827451888.0
        self._gr_k_134 = -88208028288.0*nu - 6974032128.0*tmp_init_6 + 83031623984.0
        self._gr_k_135 = -40328524800.0*nu - 3315689472.0*tmp_init_6 + 37896423296.0
        self._gr_k_136 = 67052371968.0*nu + 6901075968.0*tmp_init_6 - 72723761408.0
        self._gr_k_137 = -29067706368.0*nu - 3249340416.0*tmp_init_6 + 34478587904.0
        self._gr_k_138 = 6.4074867987318384e-10*flagPN2
        self._gr_k_139 = -158371936320.0*nu + 2669430720.0*tmp_init_6 - 450495323526.0
        self._gr_k_140 = 9437550096.0*nu - 19617275232.0*tmp_init_6 - 18495592281.0
        self._gr_k_141 = -123750355344.0*nu + 377798072160.0*tmp_init_6 + 3714206327561.0
        self._gr_k_142 = 810176372928.0*nu - 203825092032.0*tmp_init_6 + 2118647643574.0
        self._gr_k_143 = 342082738920.0*nu + 57658973472.0*tmp_init_6 + 763636538570.0
        self._gr_k_144 = 26557627080.0*nu + 20609716896.0*tmp_init_6 + 52895194690.0
        self._gr_k_145 = 381212180688.0*nu - 2641584576.0*tmp_init_6 + 2234076675540.0
        self._gr_k_146 = -64636028928.0*nu + 256331713536.0*tmp_init_6 + 2716188116928.0
        self._gr_k_147 = -75536177664.0*nu - 9396380928.0*tmp_init_6 - 110929094864.0
        self._gr_k_148 = 531839783424.0*nu - 115370094336.0*tmp_init_6 + 1256023561872.0
        self._gr_k_149 = 154838776704.0*nu + 49031444736.0*tmp_init_6 + 204687477808.0
        self._gr_k_150 = 199204592256.0*nu - 47865599232.0*tmp_init_6 + 1702626162320.0
        self._gr_k_151 = 4856472576.0*nu + 103465420800.0*tmp_init_6 + 1581970949504.0
        self._gr_k_152 = 312940127232.0*nu - 17849653248.0*tmp_init_6 + 529747731328.0
        self._gr_k_153 = -19863244800.0*nu - 36076916736.0*tmp_init_6 + 1129213141760.0
        self._gr_k_154 = 1.6018716996829596e-10*flagPN2
        self._gr_k_155 = 120700.0*nu - 1693512.0
        self._gr_k_156 = 11142264.0*nu + 11602964.0
        self._gr_k_157 = 21357444.0*nu + 42487832.0
        self._gr_k_158 = -8457592.0*nu - 21838148.0
        self._gr_k_159 = -15840100.0*nu - 24115325.0
        self._gr_k_160 = 34397880.0*nu + 51144550.0
        self._gr_k_161 = -2123412.0*nu - 1077113.0
        self._gr_k_162 = 24686336.0*nu + 46681792.0
        self._gr_k_163 = 4917056.0*nu + 7281232.0
        self._gr_k_164 = -5994560.0*nu - 24476560.0
        self._gr_k_165 = 28002560.0*nu + 50010336.0
        self._gr_k_166 = -8317184.0*nu - 19089248.0
        self._gr_k_167 = 17106944.0*nu + 40461056.0
        self._gr_k_168 = -839680.0*nu - 23030528.0
        self._gr_k_169 = 31847424.0*nu + 51537152.0
        self._gr_k_170 = 4.8440600198412698e-7*flagPN1
        self._gr_k_171 = 1199088.0*nu + 11238028.0
        self._gr_k_172 = 529772.0*nu - 3795.0
        self._gr_k_173 = 1316340.0*nu + 3087451.0
        self._gr_k_174 = -4788368.0*nu - 14413828.0
        self._gr_k_175 = 4055904.0*nu + 16511336.0
        self._gr_k_176 = -1422992.0*nu - 10734940.0
        self._gr_k_177 = -738768.0*nu - 2027212.0
        self._gr_k_178 = 1106560.0*nu + 4251872.0
        self._gr_k_179 = 1357312.0*nu + 4893440.0
        self._gr_k_180 = -5086336.0*nu - 10886624.0
        self._gr_k_181 = 4408192.0*nu + 10799904.0
        self._gr_k_182 = -1869184.0*nu - 5307040.0
        self._gr_k_183 = 2685952.0*nu + 4135168.0
        self._gr_k_184 = -5371904.0*nu - 7052800.0
        self._gr_k_185 = 1572864.0*nu + 4980736.0
        self._gr_k_186 = 9.6881200396825397e-7*flagPN1
        self._gr_k_187 = -313040.0*nu + 113120189952.0*tmp_init_1 + 47127579408.0*tmp_init_2 - 48105750000.0*tmp_init_3 - 27117623904.0*tmp_init_5 - 1717718.2178518874
        self._gr_k_188 = 7413440.0*nu - 133702636032.0*tmp_init_1 - 48578736528.0*tmp_init_2 + 71426250000.0*tmp_init_3 + 16049205984.0*tmp_init_5 - 44982216.975865189
        self._gr_k_189 = 11823600.0*nu + 26699097600.0*tmp_init_1 + 9232891920.0*tmp_init_2 - 17949750000.0*tmp_init_3 + 79683509.642912432
        self._gr_k_190 = -3696000.0*nu - 351240960.0*tmp_init_1 - 1558388160.0*tmp_init_2 + 1806000000.0*tmp_init_3 - 186732047.44608597
        self._gr_k_191 = -7658560.0*nu + 43526931840.0*tmp_init_1 + 1373591520.0*tmp_init_2 - 19845000000.0*tmp_init_3 + 45032986.972451945
        self._gr_k_192 = -711360.0*nu - 41720391552.0*tmp_init_1 - 9537411744.0*tmp_init_2 + 20454000000.0*tmp_init_3 + 3320525376.0*tmp_init_5 + 11355555.9032376
        self._gr_k_193 = 16399971840.0*tmp_init_1 - 1187706240.0*tmp_init_2 - 6216000000.0*tmp_init_3 + tmp_init_53 - 22541878.48451935
        self._gr_k_194 = 9031680.0*nu + 5712000000.0*tmp_init_3
        self._gr_k_195 = -931840.0*nu + 2239856640.0*tmp_init_1 - 888330240.0*tmp_init_2 - 126180364.85028249
        self._gr_k_196 = -1730560.0*nu - 5607383040.0*tmp_init_1 + 1924715520.0*tmp_init_2 - tmp_init_54 + 33034344.44578211
        self._gr_k_197 = -1741824000.0*tmp_init_2 + tmp_init_53
        self._gr_k_198 = -1327819259289600.0*nu - 2.3787778167298253e+19*tmp_init_1 - 6.63513421076352e+18*tmp_init_2 + 1.495137032928e+19*tmp_init_3 - 69288077576504518.0
        self._gr_k_199 = 427173626880000.0*nu - 8.8226312012653732e+19*tmp_init_1 - 4.0149840175803832e+19*tmp_init_2 + 3.540201432768e+19*tmp_init_3 + 2.4819003937302221e+19*tmp_init_5 - 3030742254318027.2
        self._gr_k_200 = -3240779209113600.0*nu + 9.6667465632366182e+19*tmp_init_1 + 4.0314722990026383e+19*tmp_init_2 - 5.241297178656e+19*tmp_init_3 - 1.3870807807950213e+19*tmp_init_5 + 24182735650656445.0
        self._gr_k_201 = 1368457755033600.0*nu - 2.9802159576870912e+18*tmp_init_1 + 1.2807488699292672e+18*tmp_init_2 + 1.408264704e+17*tmp_init_3 + 1.2447936182725777e+17
        self._gr_k_202 = 87178291200000.0*nu + 2.7210357670499942e+19*tmp_init_1 + 8.165318727379968e+18*tmp_init_2 - 1.393125858432e+19*tmp_init_3 - 2.7834396939030528e+18*tmp_init_5 - 717789664605422.89
        self._gr_k_203 = 3034609256448000.0*nu - 2.7414459163968307e+19*tmp_init_1 - 1.7309809536509952e+18*tmp_init_2 + 1.306517579136e+19*tmp_init_3 - 28888099686738391.0
        self._gr_k_204 = -1461698290483200.0*nu + 4.2370831931129856e+18*tmp_init_1 - 7.928282428932096e+17*tmp_init_2 - 1.44699198336e+18*tmp_init_3 + 86240798953057151.0
        self._gr_k_205 = -945066324787200.0*nu + 6.3617693302259712e+18*tmp_init_1 - 1.6524555504500736e+18*tmp_init_2 - 1.56317382144e+18*tmp_init_3 - 16912202094533487.0
        self._gr_k_206 = -610945464729600.0*nu - 9.1102513877286912e+18*tmp_init_1 + 2.29963993104384e+17*tmp_init_2 + 3.76006675968e+18*tmp_init_3 + 7068680937463553.4
        self._gr_k_207 = 354292575436800.0*nu - 4.022815155093504e+17*tmp_init_1 + 92350619453030400.0*tmp_init_2 + 53702916414131996.0
        self._gr_k_208 = 680688097689600.0*nu + 2.5244395358257152e+18*tmp_init_1 - 8.154596199923712e+17*tmp_init_2 - 5.0697529344e+17*tmp_init_3 - 11527832477184461.0
        self._gr_k_209 = -189699961651200.0*nu - 9.440826317733888e+17*tmp_init_1 + 4.9277998522368e+17*tmp_init_2 + 40473978564074916.0
        self._gr_k_210 = -267811710566400.0*nu - 4.451648656637952e+17*tmp_init_1 + 2.77416880570368e+17*tmp_init_2 + 4703121387615818.2
        self._gr_k_211 = tmp_init_55*(672377720.0*nu + 469966565880.0*tmp_init_1 + 265015170630.0*tmp_init_2 - 372530812500.0*tmp_init_3 - 57028513.754644763)
        self._gr_k_212 = -76090186087294500.0*nu + 3.5972902756643973e+19*tmp_init_1 + 37292935680000000.0*tmp_init_11 - 1.004436797607936e+17*tmp_init_14 + 3.0173789315843712e+17*tmp_init_2 + 33533880784588800.0*tmp_init_24 + 27796102825780800.0*tmp_init_25 + 8843580452044800.0*tmp_init_26 - 1.165727937674539e+20*tmp_init_27 + 8.332232832e+18*tmp_init_28 - 1.601699920128e+19*tmp_init_3 + 1.392597759168e+19*tmp_init_30 + 4.068431064525865e+19*tmp_init_33 + 1.8678186551033856e+19*tmp_init_34 + 5.0154760656e+18*tmp_init_35 + 180433915200.0*tmp_init_37*(79111.0*nu - 178800.0) - tmp_init_56*(990020.0*nu + 17736130860.0*tmp_init_1 + 4947145875.0*tmp_init_2 - 11147718750.0*tmp_init_3 + 40036914.0) + 1.357862939955188e+17*tmp_init_6 - 1.4369488240406477e+17
        self._gr_k_213 = tmp_init_38*(62283200.0*nu - 636449314536.0*tmp_init_1 - 336978906453.0*tmp_init_2 + 357549609375.0*tmp_init_3 + 121251059433.0*tmp_init_5 - 152645424.69785172)
        self._gr_k_214 = 8586925839874500.0*nu + 4.9805886178009995e+19*tmp_init_1 - 21552410880000000.0*tmp_init_11 + 72466082777088000.0*tmp_init_14 + 2.0528903862125455e+19*tmp_init_2 + 1003529428070400.0*tmp_init_24 - 760695261249600.0*tmp_init_25 - 13451261618995200.0*tmp_init_26 - 2.658275409526965e+20*tmp_init_27 - 7.393389696e+19*tmp_init_28 + 2.3133476567105372e+19*tmp_init_29 - 2.020476349584e+19*tmp_init_3 + 7.11349708608e+19*tmp_init_30 + tmp_init_32 + 1.1231390139850654e+20*tmp_init_33 - 1.7582736643881615e+19*tmp_init_34 + 1.71089492112e+19*tmp_init_35 + 1.0607997944541635e+19*tmp_init_36 - 1.2630115336982686e+19*tmp_init_5 + tmp_init_56*(318500.0*nu - 65781402708.0*tmp_init_1 - 29935659159.0*tmp_init_2 + 26395687500.0*tmp_init_3 + 18505011210.0*tmp_init_5 - 3890190.0) - tmp_init_57*(162975.0*nu - 538384.0) + 1910129785930800.0*tmp_init_6 + 2771790205979205.0
        self._gr_k_215 = -tmp_init_55*(6215080.0*nu - 2692360603800.0*tmp_init_1 - 1297446653034.0*tmp_init_2 + 1029829500000.0*tmp_init_3 + 839885387292.0*tmp_init_5 - 28279334.475758088)
        self._gr_k_216 = 3678226462310400.0*nu*tmp_init_2 + 7507019520000000.0*nu*tmp_init_3 + 1.106938522061844e+17*nu + 3.8827003788636586e+20*tmp_init_1*tmp_init_2 + 4.2717362688e+19*tmp_init_1*tmp_init_3 - 6.4040904753415705e+19*tmp_init_1 - 31221738145382400.0*tmp_init_14 - 1.8077601555946437e+19*tmp_init_2 - 9153157270118400.0*tmp_init_24 + 768269452204800.0*tmp_init_25 - 1.2123426222333297e+19*tmp_init_29 + 3.196614882384e+19*tmp_init_3 - 8.141706353088e+19*tmp_init_30 - 2.5050957245127475e+18*tmp_init_31 - 1.7686430426189758e+20*tmp_init_33 - 1.4268261961046016e+19*tmp_init_34 - 1.9957751352e+19*tmp_init_35 - 6.2782028651368858e+18*tmp_init_36 - 293388480.0*tmp_init_37*(21179985.0*nu - 4805584.0) + 6.7024196925595177e+18*tmp_init_5 - tmp_init_58*(402720.0*nu - 12012518980.0*tmp_init_1 - 5009765922.0*tmp_init_2 + 6513171875.0*tmp_init_3 + 1723675499.0*tmp_init_5 - 8222072.0) - 41824776896596800.0*tmp_init_6 - 2.8808135537566115e+17
        self._gr_k_217 = -8.948693657085996e+17*nu - 1.2490118656114176e+19*tmp_init_1 - 9202152960000000.0*tmp_init_11 + 15318558026956800.0*tmp_init_14 + 1.2952923162689664e+18*tmp_init_2 - 36956808974131200.0*tmp_init_24 - 16155752565024000.0*tmp_init_25 + 5808971195596800.0*tmp_init_26 - 6.805534944079872e+18*tmp_init_27 + 5.00168680704e+18*tmp_init_28 + 5.00341919616e+18*tmp_init_3 + 7.4461996224e+17*tmp_init_30 - 4.5028764445851648e+18*tmp_init_33 + 2.854473692184576e+18*tmp_init_34 - 1.3774589136e+18*tmp_init_35 + 10268596800.0*tmp_init_37*(3544983.0*nu + 2890480.0) + tmp_init_43*(510160.0*nu - 1111022220.0*tmp_init_1 + 477462195.0*tmp_init_2 + 52500000.0*tmp_init_3 + 93011832.0) + 74686819431280320.0*tmp_init_6 - 8.2412257776361398e+17
        self._gr_k_218 = tmp_init_44*(-2176276.0*nu - 611043048.0*tmp_init_1 - 31953285.0*tmp_init_2 + 534909375.0*tmp_init_3 + 16725763.917404265)
        self._gr_k_219 = -40300201450260808.0*nu - 4.2189281964166204e+18*tmp_init_1 + 2905943040000000.0*tmp_init_11 - 21543034370457600.0*tmp_init_14 - 6.3573508694535294e+17*tmp_init_2 + 359389152460800.0*tmp_init_24 + 1298456813030400.0*tmp_init_25 + 9327379732070400.0*tmp_init_26 + 8.2953633654804972e+19*tmp_init_27 + 2.03728960512e+18*tmp_init_28 - 2.9690023401632563e+18*tmp_init_29 + 1.9723372956110292e+18*tmp_init_3 - 1.432205203968e+19*tmp_init_30 - 1.8714436411126579e+19*tmp_init_33 - 6.1409785480810906e+18*tmp_init_34 - 5.7152075904e+18*tmp_init_35 - 1.2989385238214246e+18*tmp_init_36 + 2.3402728294371828e+17*tmp_init_5 + tmp_init_57*(1641845.0*nu - 498176.0) + 346242193776000.0*tmp_init_6 + 4204129306173104.1
        self._gr_k_220 = -379904840682758.54*nu - 2.3021800212564312e+19*tmp_init_1 - 8.6139435063607029e+18*tmp_init_2 + 1.2273912509472513e+19*tmp_init_3 + 2.9148112313586845e+18*tmp_init_5 - 2172361405992151.3
        self._gr_k_221 = -7.8150256110093001e+17*nu + 4.5995183109036038e+18*tmp_init_1 + 8224408933171200.0*tmp_init_14 - 4.4736693170563743e+17*tmp_init_2 + 9241596293529600.0*tmp_init_24 + 7635966161817600.0*tmp_init_25 - 534979642982400.0*tmp_init_26 - 7.9563404695722394e+19*tmp_init_27 + 3.32350470144e+18*tmp_init_28 - 1.6610544793964846e+18*tmp_init_3 + 1.032962160384e+19*tmp_init_30 + 1.3151299627597824e+19*tmp_init_33 + 1.4996526900319642e+19*tmp_init_34 + 5.545042272e+18*tmp_init_35 + 4400827200.0*tmp_init_37*(7757733.0*nu + 259808.0) + 1.241057806506144e+17*tmp_init_6 - 5.011117662507344e+17
        self._gr_k_222 = -7423455850070701.4*nu + 1.9900614132333375e+19*tmp_init_1 + 3.6701012333321926e+18*tmp_init_2 - 1.1124082910436207e+19*tmp_init_3 - 29089275483430919.0
        self._gr_k_223 = 32284043624448000.0*nu*tmp_init_1 + 1.1332042716403056e+18*nu + 1.0959670219515494e+19*tmp_init_1*tmp_init_2 - 3.1168960193297203e+18*tmp_init_1 + 3.168595584e+17*tmp_init_2*tmp_init_3 + 1.0869133304174746e+18*tmp_init_2 - 18055080117043200.0*tmp_init_24 - 455355194342400.0*tmp_init_25 - 15488336631398400.0*tmp_init_26 - 3.5676039168e+18*tmp_init_28 + 8.523633888e+17*tmp_init_3 + 3.0625994428710912e+18*tmp_init_33 - 4.5570285411323904e+18*tmp_init_34 - 5.046281856e+17*tmp_init_35 - 8801654400.0*tmp_init_37*(4315619.0*nu + 248240.0) - tmp_init_42 - tmp_init_58*(181640.0*nu - 526527120.0*tmp_init_1 + 98521920.0*tmp_init_2 + 179812500.0*tmp_init_3 - 23494797.0) - 1.1841114386018816e+17*tmp_init_6 - 3.9014246716570541e+17
        self._gr_k_224 = tmp_init_45*(286450.0*nu + 2701010760.0*tmp_init_1 + 124411140.0*tmp_init_2 - 1248843750.0*tmp_init_3 - 1650591.1077204487)
        self._gr_k_225 = -35020935096046728.0*nu - 4.1204567978814173e+18*tmp_init_1 - 2648932496179200.0*tmp_init_14 + 2.4702335062874416e+18*tmp_init_2 + 11338220784844800.0*tmp_init_24 + 12352648319539200.0*tmp_init_25 + 853649827430400.0*tmp_init_26 + 1.139063977696297e+19*tmp_init_27 - 6.0085960704e+17*tmp_init_28 + 61588237187226931.0*tmp_init_3 - 7.6046294016e+17*tmp_init_30 + 2.4062985198895104e+18*tmp_init_33 - 4.7310528803586048e+18*tmp_init_34 - 7.980166656e+17*tmp_init_35 + 23471078400.0*tmp_init_37*(392001.0*nu - 681848.0) + 46136135198476800.0*tmp_init_6 - 2.8526556984740032e+17
        self._gr_k_226 = 4163349185326952.9*nu - 4.2619452851131008e+18*tmp_init_1 + 6.531225312332994e+17*tmp_init_2 + 1.5355640176308494e+18*tmp_init_3 - 2258518028859181.9
        self._gr_k_227 = 3793999233024000.0*nu*tmp_init_1 + 12188319122520000.0*nu + 2.4409921536e+18*tmp_init_1*tmp_init_3 + 5.4662185580986368e+18*tmp_init_1 + 1.52092588032e+18*tmp_init_2*tmp_init_3 - 7.8761190167396352e+17*tmp_init_2 - 2784984278630400.0*tmp_init_24 - 86256909388800.0*tmp_init_25 - 3263955222528000.0*tmp_init_26 - 1.7763927585855898e+19*tmp_init_27 - 1.78123131648e+18*tmp_init_3 - 3.7338978216443904e+18*tmp_init_33 + 4.4863054876975104e+18*tmp_init_34 + 1.7368598016e+18*tmp_init_35 - 164297548800.0*tmp_init_37*(3813.0*nu - 5080.0) - tmp_init_46*(9490.0*nu + 141512280.0*tmp_init_1 - 3572100.0*tmp_init_2 - 58406250.0*tmp_init_3 - 313539.0) - 11579725119916800.0*tmp_init_6 - 91568537037364752.0
        self._gr_k_228 = 5322757747507200.0*nu*tmp_init_2 - 4.1329146106209571e+17*nu + 5.105443995058176e+17*tmp_init_1*tmp_init_2 + 2.5702905247891834e+18*tmp_init_1 - 7565680823500800.0*tmp_init_14 - 1.4182257591381409e+18*tmp_init_2 - 15979861249228800.0*tmp_init_24 - 5078274477235200.0*tmp_init_25 - 1.2759013411651584e+18*tmp_init_33 + 182511105638400.0*tmp_init_34 + 281652940800.0*tmp_init_37*(51209.0*nu + 54232.0) + 49238305128695040.0*tmp_init_6 - 2.098567646076583e+17
        self._gr_k_229 = tmp_init_59*(-34900.0*nu + 50871240.0*tmp_init_1 - 13919850.0*tmp_init_2 + 1782234.1909384632)
        self._gr_k_230 = -2.568457785804912e+17*nu - 1.193105934531625e+18*tmp_init_1 + 624893991321600.0*tmp_init_14 + 4.676941234538496e+17*tmp_init_2 + 3803441312563200.0*tmp_init_24 + 2674851514444800.0*tmp_init_25 + 602576348774400.0*tmp_init_26 + 2.1784525568999424e+18*tmp_init_27 - 4.5064470528e+17*tmp_init_28 + 1.9179224064e+17*tmp_init_3 + 2.9438275243474944e+18*tmp_init_33 - 1.4922107996995584e+18*tmp_init_34 - 2.816529408e+17*tmp_init_35 + 2886942643200.0*tmp_init_37*(3641.0*nu + 96.0) + 42617119726483200.0*tmp_init_6 + tmp_init_60*(31720.0*nu + 117638640.0*tmp_init_1 - 38000340.0*tmp_init_2 - tmp_init_49 - 124431.0) - 1.4537534742723845e+17
        self._gr_k_231 = -tmp_init_59*(42900.0*nu + 126005880.0*tmp_init_1 - 32755590.0*tmp_init_2 - 31500000.0*tmp_init_3 + 938827.81648380006)
        self._gr_k_232 = 892705701888000.0*nu*tmp_init_1 + 4.97066948404536e+17*nu + 3.1751082583621632e+17*tmp_init_1 - 33679817244057600.0*tmp_init_2 - 8634352553164800.0*tmp_init_24 + 414199095321600.0*tmp_init_25 - 1.5827363080962048e+18*tmp_init_27 - 3.743595695702016e+17*tmp_init_33 + 1.0162218361946112e+18*tmp_init_34 - 70413235200.0*tmp_init_37*(209469.0*nu + 20320.0) - tmp_init_48 - 52907117054227200.0*tmp_init_6 - tmp_init_60*(8840.0*nu + 43994160.0*tmp_init_1 - 22963500.0*tmp_init_2 - 4166643.0) - 2.4854217841495426e+17
        self._gr_k_233 = -20173203007695334.0*nu - tmp_init_1*tmp_init_41 - 64137859906753588.0*tmp_init_1 + 83928015971050892.0*tmp_init_2 + 648928375603200.0*tmp_init_24 + 2727980120678400.0*tmp_init_25 - 70084264565145600.0*tmp_init_27 - 6.852683646369792e+17*tmp_init_33 + 3.036984797822976e+17*tmp_init_34 + 1126611763200.0*tmp_init_37*(4305.0*nu - 4168.0) + 2995819638220800.0*tmp_init_6 - 1.8770366575286919e+17
        self._gr_k_234 = 345171406137730.56*nu + 4.1300028409540507e+17*tmp_init_1 - 2.0144594276681471e+17*tmp_init_2 - 1663341656358832.3
        self._gr_k_235 = -1.269289677429504e+17*nu + 417541129947762.83*tmp_init_1 + 31422358788814974.0*tmp_init_2 - 4109879712153600.0*tmp_init_24 + 1044361259827200.0*tmp_init_25 - 60566648389632000.0*tmp_init_33 - 1.080465745379328e+17*tmp_init_34 + 563305881600.0*tmp_init_37*(5453.0*nu + 9456.0) + tmp_init_50 + 20449278555033600.0*tmp_init_6 - 1283386066742351.6
        self._gr_k_236 = 1.316421784485888e+17*nu - 13638825797148276.0*tmp_init_1 + 57645698165637120.0*tmp_init_2 - 2595713502412800.0*tmp_init_24 + 708329373696000.0*tmp_init_25 + 20765708019302400.0*tmp_init_33 - 35042132282572800.0*tmp_init_34 - 25912070553600.0*tmp_init_37*(123.0*nu + 32.0) - 1189116266023.0881*tmp_init_52 - 10430846912102400.0*tmp_init_6 - 94365531598066954.0
        self._gr_k_237 = tmp_init_61
        self._gr_k_238 = -2.3115031741456849e-15*flagPN3
        self._gr_k_239 = -6744600.0*nu + 148830023160.0*tmp_init_1 + 5827813110.0*tmp_init_2 - 67605562500.0*tmp_init_3 - 133655189.13941998
        self._gr_k_240 = -4619680.0*nu - 155322000288.0*tmp_init_1 - 59338534500.0*tmp_init_2 + 79651687500.0*tmp_init_3 + 22897789572.0*tmp_init_5 + 29992375.483625689
        self._gr_k_241 = -336440.0*nu + 50797878936.0*tmp_init_1 + 27204273306.0*tmp_init_2 - 15944250000.0*tmp_init_3 - 20269040316.0*tmp_init_5 + 4333430.750332009
        self._gr_k_242 = 3633440.0*nu + 58420958400.0*tmp_init_1 + 2986547760.0*tmp_init_2 - 27058500000.0*tmp_init_3 - 34193592.70117969
        self._gr_k_243 = 580320.0*nu - 16401669312.0*tmp_init_1 - 9443516544.0*tmp_init_2 + 9507750000.0*tmp_init_3 + 3320525376.0*tmp_init_5 - 7911657.8014360331
        self._gr_k_244 = 3109440.0*nu - 43062902400.0*tmp_init_1 + 13233371760.0*tmp_init_2 + 9570750000.0*tmp_init_3 + 13250969.109143131
        self._gr_k_245 = -1200000.0*nu - 19223877120.0*tmp_init_1 + 5366450880.0*tmp_init_2 + 4935000000.0*tmp_init_3 - 90641367.239553782
        self._gr_k_246 = -486720.0*nu + 7198060800.0*tmp_init_1 + 1480550400.0*tmp_init_2 - 4137000000.0*tmp_init_3 + 10458619.29687159
        self._gr_k_247 = 715520.0*nu + 5216924160.0*tmp_init_1 - 3132017280.0*tmp_init_2 - 15907932.263849992
        self._gr_k_248 = -83200.0*nu - 3001904640.0*tmp_init_1 + 517104000.0*tmp_init_2 - tmp_init_54 - 12658652.482297653
        self._gr_k_249 = 798720.0*nu + 1155409920.0*tmp_init_1 - 439810560.0*tmp_init_2 - 67422555.467211029
        self._gr_k_250 = tmp_init_61


    cdef void _compute(self
            , double e=-1
            , double x=-1
            , double z=-1
        ):
        """
        values being computed:
        - FphiCorrMultParser
        - FrCorrMultParser
        """

        # internal computations intermediate variables declaration/initialisation
        cdef:
            double tmp_0 = cmath.cos(z)
            double tmp_1 = e*tmp_0
            double tmp_2 = cmath.pow(e, 2)
            double tmp_3 = 2.0*z
            double tmp_4 = cmath.cos(tmp_3)
            double tmp_5 = cmath.pow(e, 3)
            double tmp_6 = 3.0*z
            double tmp_7 = cmath.cos(tmp_6)
            double tmp_8 = cmath.pow(e, 4)
            double tmp_9 = 4.0*z
            double tmp_10 = cmath.cos(tmp_9)
            double tmp_11 = cmath.pow(e, 5)
            double tmp_12 = 5.0*z
            double tmp_13 = cmath.cos(tmp_12)
            double tmp_14 = cmath.pow(e, 6)
            double tmp_15 = 6.0*z
            double tmp_16 = cmath.cos(tmp_15)
            double tmp_17 = e*x
            double tmp_18 = e*cmath.pow(x, 2)
            double tmp_19 = cmath.pow(x, 2.5)
            double tmp_20 = cmath.sin(z)
            double tmp_21 = cmath.sin(tmp_3)
            double tmp_22 = tmp_0*tmp_20
            double tmp_23 = cmath.sin(tmp_6)
            double tmp_24 = cmath.sin(tmp_9)
            double tmp_25 = cmath.sin(tmp_12)
            double tmp_26 = cmath.sin(tmp_15)
            double tmp_27 = cmath.log(x)
            double tmp_28 = tmp_2*tmp_21
            double tmp_29 = cmath.pow(x, 1.5)
            double tmp_30 = 3.1415926535897932*tmp_0
            double tmp_31 = 1.0986122886681097
            double tmp_32 = 0.69314718055994531
            double tmp_33 = 10.073412021181606
            double tmp_34 = 3870720.0*tmp_20*(54.0*tmp_31 - 24.0*tmp_32 - tmp_33)
            double tmp_35 = 3.1415926535897932*tmp_7
            double tmp_36 = 3.1415926535897932*tmp_4
            double tmp_37 = 3.1415926535897932*tmp_13
            double tmp_38 = 3.1415926535897932*tmp_10
            double tmp_39 = 3.1415926535897932*tmp_16
            double tmp_40 = 69672960.0*tmp_20
            double tmp_41 = 3.1415926535897932*tmp_20
            double tmp_42 = tmp_41*(945.0*tmp_31 + 420.0*tmp_32 + 289.4305792586438)
            double tmp_43 = cmath.pow(tmp_27, 2)
            double tmp_44 = 15120.0*tmp_32
            double tmp_45 = 3.1415926535897932*tmp_23
            double tmp_46 = 3.1415926535897932*tmp_21
            double tmp_47 = 3.1415926535897932*tmp_24
            double tmp_48 = 3.1415926535897932*tmp_25
            double tmp_49 = 3.1415926535897932*tmp_26
            double tmp_50 = e*cmath.pow(x, 3)
            double tmp_51 = e*tmp_21

        # azimuthal/math.m: Azimuthal eccentricity correction to the RR force, computed in an eccentricity expansion, up to O(e^6). This corresponds to Eq. (57a) of [docs/SEOBNRv5E_theory.pdf]
        self.FphiCorrMultParser = self._gr_k_138*tmp_18*(self._gr_k_137*tmp_0 + e*(self._gr_k_135*tmp_4 + self._gr_k_136) + tmp_11*(self._gr_k_123*tmp_10 + self._gr_k_124*tmp_16 + self._gr_k_125*tmp_4 + self._gr_k_126) + tmp_2*(self._gr_k_133*tmp_7 + self._gr_k_134*tmp_0) + tmp_5*(self._gr_k_130*tmp_4 + self._gr_k_131*tmp_10 + self._gr_k_132) + tmp_8*(self._gr_k_127*tmp_0 + self._gr_k_128*tmp_7 + self._gr_k_129*tmp_13)) + self._gr_k_186*tmp_17*(self._gr_k_185*tmp_0 + e*(self._gr_k_183*tmp_4 + self._gr_k_184) + tmp_11*(self._gr_k_171*tmp_10 + self._gr_k_172*tmp_16 + self._gr_k_173*tmp_4 + self._gr_k_174) + tmp_2*(self._gr_k_181*tmp_0 + self._gr_k_182*tmp_7) + tmp_5*(self._gr_k_178*tmp_10 + self._gr_k_179*tmp_4 + self._gr_k_180) + tmp_8*(self._gr_k_175*tmp_0 + self._gr_k_176*tmp_7 + self._gr_k_177*tmp_13)) + self._gr_k_237*tmp_29*(e*(-14353920.0*tmp_30 + tmp_34) + tmp_11*(self._gr_k_190*tmp_20 + self._gr_k_191*tmp_23 + self._gr_k_192*tmp_25 + 13573140.0*tmp_30 - 27486270.0*tmp_35 + 320586.0*tmp_37) + tmp_14*(self._gr_k_187*tmp_26 + self._gr_k_188*tmp_24 + self._gr_k_189*tmp_21 + 19187595.0*tmp_36 + 12557874.0*tmp_38 - 1717611.0*tmp_39 - 202966646.03969172) + tmp_2*(self._gr_k_197*tmp_0*tmp_20 - 1290240.0*tmp_21*(-1044.0*tmp_32 - tmp_33) - 2499840.0*tmp_36 - 57001057.106733209) + tmp_27*(e*tmp_40 + tmp_11*(333668160.0*tmp_20 - 80468640.0*tmp_23 - 20291040.0*tmp_25) + tmp_14*(-142385040.0*tmp_21 + 80377920.0*tmp_24 + 3069360.0*tmp_26) - 23224320.0*tmp_28 + tmp_5*(225469440.0*tmp_20 - 59028480.0*tmp_23) + tmp_8*(-75720960.0*tmp_21 + 40279680.0*tmp_24)) + tmp_5*(self._gr_k_195*tmp_20 + self._gr_k_196*tmp_23 - 90720.0*tmp_30 - 9787680.0*tmp_35) + tmp_8*(self._gr_k_193*tmp_24 + self._gr_k_194*tmp_22 - 13440.0*tmp_21*(-247050.0*tmp_31 + 896504.0*tmp_32 - 3152.9779626298426) + 8057280.0*tmp_36 + 3158400.0*tmp_38 - 132305786.99551741)) + self._gr_k_238*tmp_50*(self._gr_k_236*tmp_0 + e*(self._gr_k_233*tmp_4 + self._gr_k_234*tmp_21 + self._gr_k_235) + tmp_11*(self._gr_k_211*tmp_21 + self._gr_k_212*tmp_4 + self._gr_k_213*tmp_24 + self._gr_k_214*tmp_16 + self._gr_k_215*tmp_26 + self._gr_k_216*tmp_10 + self._gr_k_217) + tmp_2*(self._gr_k_229*tmp_20 + self._gr_k_230*tmp_7 + self._gr_k_231*tmp_23 + self._gr_k_232*tmp_0) + tmp_27*(e*(self._gr_k_210*tmp_4 - 55483376114073600.0*tmp_31 + 24659278272921600.0*tmp_32 + 6854305967308800.0*tmp_46 + 13491250275936298.0) - 1030045040640.0*tmp_0*(34020.0*tmp_31 - tmp_44 - 11375.249573344412) + tmp_11*(self._gr_k_198*tmp_4 + self._gr_k_199*tmp_16 + self._gr_k_200*tmp_10 + self._gr_k_201 + 37884767397523200.0*tmp_46 + 4548849903037440.0*tmp_47 - 1295339478048000.0*tmp_49) + tmp_2*(self._gr_k_208*tmp_7 + self._gr_k_209*tmp_0 + 45049824575078400.0*tmp_41 - 3538687548211200.0*tmp_45) + 7787140507238400.0*tmp_41 + tmp_5*(self._gr_k_205*tmp_4 + self._gr_k_206*tmp_10 + self._gr_k_207 + 20032283761459200.0*tmp_46 + 410649987686400.0*tmp_47) + tmp_8*(self._gr_k_202*tmp_13 + self._gr_k_203*tmp_7 + self._gr_k_204*tmp_0 + 1.147141446054912e+17*tmp_41 - 16089283416729600.0*tmp_45 + 1239977071872000.0*tmp_48)) + 12360540487680.0*tmp_42 + tmp_43*(e*(1460088845107200.0*tmp_4 - 9247229352345600.0) - 5840355380428800.0*tmp_0 + tmp_11*(-20594603857766400.0*tmp_10 + 2257941213158400.0*tmp_16 + 75451231765324800.0*tmp_4 - 83152820191795200.0) + tmp_2*(-19427293244620800.0*tmp_0 + 8557742953267200.0*tmp_7) + tmp_5*(-6266214626918400.0*tmp_10 + 25510996765900800.0*tmp_4 - 35954687810764800.0) + tmp_8*(-40623930263347200.0*tmp_0 + 808625593036800.0*tmp_13 + 20793591660441600.0*tmp_7)) + tmp_5*(self._gr_k_224*tmp_24 + self._gr_k_225*tmp_4 + self._gr_k_226*tmp_21 + self._gr_k_227*tmp_10 + self._gr_k_228) + tmp_8*(self._gr_k_218*tmp_20 + self._gr_k_219*tmp_13 + self._gr_k_220*tmp_25 + self._gr_k_221*tmp_7 + self._gr_k_222*tmp_23 + self._gr_k_223*tmp_0)) + self._gr_k_42*tmp_19*(e*(self._gr_k_40*tmp_20 + self._gr_k_41*tmp_0) + tmp_11*(self._gr_k_20*tmp_0 + self._gr_k_21*tmp_13 + self._gr_k_22*tmp_20 + self._gr_k_23*tmp_25 + self._gr_k_24*tmp_7 + self._gr_k_25*tmp_23) + tmp_14*(self._gr_k_12*tmp_4 + self._gr_k_13*tmp_10 + self._gr_k_14*tmp_21 + self._gr_k_15*tmp_22 + self._gr_k_16*tmp_16 + self._gr_k_17*tmp_26 + self._gr_k_18*tmp_24 + self._gr_k_19) + tmp_2*(self._gr_k_36*tmp_4 + self._gr_k_37*tmp_21 + self._gr_k_38*tmp_22 + self._gr_k_39) + tmp_27*(self._gr_k_10*e*tmp_20 + self._gr_k_11*tmp_28 + tmp_11*(self._gr_k_3*tmp_20 + self._gr_k_4*tmp_23 + self._gr_k_5*tmp_25) + tmp_14*(self._gr_k_0*tmp_26 + self._gr_k_1*tmp_24 + self._gr_k_2*tmp_21) + tmp_5*(self._gr_k_8*tmp_20 + self._gr_k_9*tmp_23) + tmp_8*(self._gr_k_6*tmp_24 + self._gr_k_7*tmp_21)) + tmp_5*(self._gr_k_32*tmp_0 + self._gr_k_33*tmp_20 + self._gr_k_34*tmp_23 + self._gr_k_35*tmp_7) + tmp_8*(self._gr_k_26*tmp_21 + self._gr_k_27*tmp_4 + self._gr_k_28*tmp_24 + self._gr_k_29*tmp_10 + self._gr_k_30*tmp_22 + self._gr_k_31)) - 1.8333333333333333*tmp_1 + 0.00032552083333333333*tmp_11*(-5412.0*tmp_0 + 2190.0*tmp_13 + 1846.0*tmp_7) + 0.00032552083333333333*tmp_14*(-3416.0*tmp_10 - 1167.0*tmp_16 + 2503.0*tmp_4 + 2896.0) + 0.00032552083333333333*tmp_2*(6144.0 - 1280.0*tmp_4) + 0.00032552083333333333*tmp_5*(-6688.0*tmp_0 + 3232.0*tmp_7) + 0.00032552083333333333*tmp_8*(-3104.0*tmp_10 + 1504.0*tmp_4 + 3840.0) + 1.0
        # radial/math.m: Radial eccentricity correction to the RR force, computed in an eccentricity expansion, up to O(e^6). This corresponds to Eq. (57b) of [docs/SEOBNRv5E_theory.pdf]
        self.FrCorrMultParser = self._gr_k_122*tmp_50*(self._gr_k_121*tmp_0 + e*(self._gr_k_118*tmp_4 + self._gr_k_119*tmp_21 + self._gr_k_120) + tmp_11*(self._gr_k_83*tmp_26 + self._gr_k_84*tmp_10 + self._gr_k_85*tmp_21 + self._gr_k_86*tmp_16 + self._gr_k_87*tmp_24 + self._gr_k_88*tmp_4 + self._gr_k_89) + tmp_2*(self._gr_k_114*tmp_23 + self._gr_k_115*tmp_0 + self._gr_k_116*tmp_7 + self._gr_k_117*tmp_20) + tmp_27*(e*(self._gr_k_102*tmp_4 + 2.5551554789376e+17*tmp_31 - 1.1356246573056e+17*tmp_32 + 4866962817024000.0*tmp_46 - 23803554421272715.0) + 2060090081280.0*tmp_0*(34020.0*tmp_31 - tmp_44 + 42552.750426655588) + tmp_11*(self._gr_k_90*tmp_4 + self._gr_k_91*tmp_16 + self._gr_k_92*tmp_10 + self._gr_k_93 - 2686757815526400.0*tmp_46 - 2998927852462080.0*tmp_47 + 302424845184000.0*tmp_49) + tmp_2*(self._gr_k_100*tmp_0 + self._gr_k_101*tmp_7 - 55943033713459200.0*tmp_41 + 20846824066252800.0*tmp_45) - 15574281014476800.0*tmp_41 + tmp_5*(self._gr_k_97*tmp_4 + self._gr_k_98*tmp_10 + self._gr_k_99 - 529507528704000.0*tmp_46 - 1050002163302400.0*tmp_47) + tmp_8*(self._gr_k_94*tmp_7 + self._gr_k_95*tmp_13 + self._gr_k_96*tmp_0 - 1.4010825540096e+17*tmp_41 + 51571498419302400.0*tmp_45 + 1169704663142400.0*tmp_48)) - 24721080975360.0*tmp_42 + tmp_43*(e*(21171288254054400.0*tmp_4 + 42585924648960000.0) + 11680710760857600.0*tmp_0 + tmp_11*(-30592790363520000.0*tmp_10 - 1533917122214400.0*tmp_16 + 2.107879694895744e+17*tmp_4 + 2.679129949757184e+17) + tmp_2*(1.27554983829504e+17*tmp_0 - 1419530821632000.0*tmp_7) + tmp_5*(-10433551538995200.0*tmp_10 + 80142654386995200.0*tmp_4 + 1.40381458753536e+17) + tmp_8*(3.11485620289536e+17*tmp_0 + 4131848641536000.0*tmp_13 + 7508304095846400.0*tmp_7)) + tmp_5*(self._gr_k_109*tmp_4 + self._gr_k_110*tmp_24 + self._gr_k_111*tmp_10 + self._gr_k_112*tmp_21 + self._gr_k_113) + tmp_8*(self._gr_k_103*tmp_13 + self._gr_k_104*tmp_23 + self._gr_k_105*tmp_7 + self._gr_k_106*tmp_25 + self._gr_k_107*tmp_0 + self._gr_k_108*tmp_20)) + self._gr_k_154*tmp_18*(self._gr_k_153*tmp_0 + e*(self._gr_k_151*tmp_4 + self._gr_k_152) + tmp_11*(self._gr_k_139*tmp_10 + self._gr_k_140*tmp_16 + self._gr_k_141*tmp_4 + self._gr_k_142) + tmp_2*(self._gr_k_149*tmp_7 + self._gr_k_150*tmp_0) + tmp_5*(self._gr_k_146*tmp_4 + self._gr_k_147*tmp_10 + self._gr_k_148) + tmp_8*(self._gr_k_143*tmp_7 + self._gr_k_144*tmp_13 + self._gr_k_145*tmp_0)) + self._gr_k_170*tmp_17*(self._gr_k_169*tmp_0 + e*(self._gr_k_167*tmp_4 + self._gr_k_168) + tmp_11*(self._gr_k_155*tmp_16 + self._gr_k_156*tmp_10 + self._gr_k_157*tmp_4 + self._gr_k_158) + tmp_2*(self._gr_k_165*tmp_0 + self._gr_k_166*tmp_7) + tmp_5*(self._gr_k_162*tmp_4 + self._gr_k_163*tmp_10 + self._gr_k_164) + tmp_8*(self._gr_k_159*tmp_7 + self._gr_k_160*tmp_0 + self._gr_k_161*tmp_13)) + self._gr_k_250*e*tmp_29*(e*(self._gr_k_249*tmp_21 + 50265600.0*tmp_36 - 22124854.758465334) + tmp_11*(self._gr_k_239*tmp_21 + self._gr_k_240*tmp_24 + self._gr_k_241*tmp_26 + 84640605.0*tmp_36 + 18147570.0*tmp_38 + 1640499.0*tmp_39 - 11469483.539358297) + tmp_2*(self._gr_k_247*tmp_20 + self._gr_k_248*tmp_23 + 95074560.0*tmp_30 - 10886400.0*tmp_35) + tmp_27*(tmp_11*(238826070.0*tmp_21 - 53592840.0*tmp_24 - 7743330.0*tmp_26) + tmp_2*(28425600.0*tmp_20 + 22619520.0*tmp_23) + tmp_40 + tmp_5*(161965440.0*tmp_21 - 18688320.0*tmp_24) + 120476160.0*tmp_51 + tmp_8*(-23677920.0*tmp_20 + 61099920.0*tmp_23 + 14137200.0*tmp_25)) + 86446080.0*tmp_30 + tmp_34 + tmp_5*(self._gr_k_245*tmp_21 + self._gr_k_246*tmp_24 + 79699200.0*tmp_36 + 8032080.0*tmp_38 - 21813460.094641514) + tmp_8*(self._gr_k_242*tmp_23 + self._gr_k_243*tmp_25 + self._gr_k_244*tmp_20 + 118265280.0*tmp_30 - 19846680.0*tmp_35 - 4326504.0*tmp_37)) + self._gr_k_82*e*tmp_19*(self._gr_k_80*tmp_20 + self._gr_k_81*tmp_0 + e*(self._gr_k_77*tmp_4 + self._gr_k_78*tmp_21 + self._gr_k_79) + tmp_11*(self._gr_k_55*tmp_16 + self._gr_k_56*tmp_4 + self._gr_k_57*tmp_10 + self._gr_k_58*tmp_21 + self._gr_k_59*tmp_24 + self._gr_k_60*tmp_26 + self._gr_k_61) + tmp_2*(self._gr_k_73*tmp_20 + self._gr_k_74*tmp_7 + self._gr_k_75*tmp_0 + self._gr_k_76*tmp_23) + tmp_27*(self._gr_k_53*tmp_20 + self._gr_k_54*tmp_51 + tmp_11*(self._gr_k_43*tmp_24 + self._gr_k_44*tmp_26 + self._gr_k_45*tmp_21) + tmp_2*(self._gr_k_51*tmp_20 + self._gr_k_52*tmp_23) + tmp_5*(self._gr_k_49*tmp_24 + self._gr_k_50*tmp_21) + tmp_8*(self._gr_k_46*tmp_20 + self._gr_k_47*tmp_25 + self._gr_k_48*tmp_23)) + tmp_5*(self._gr_k_68*tmp_4 + self._gr_k_69*tmp_24 + self._gr_k_70*tmp_10 + self._gr_k_71*tmp_21 + self._gr_k_72) + tmp_8*(self._gr_k_62*tmp_7 + self._gr_k_63*tmp_20 + self._gr_k_64*tmp_13 + self._gr_k_65*tmp_23 + self._gr_k_66*tmp_0 + self._gr_k_67*tmp_25)) + 2.2916666666666667*tmp_1 + 8.1380208333333333e-5*tmp_11*(19212.0*tmp_0 - 4242.0*tmp_13 - 4634.0*tmp_7) + 8.1380208333333333e-5*tmp_14*(9490.0*tmp_10 + 2991.0*tmp_16 - 8543.0*tmp_4 - 10178.0) + 8.1380208333333333e-5*tmp_2*(-5888.0*tmp_4 - 25344.0) + 8.1380208333333333e-5*tmp_5*(26336.0*tmp_0 - 2144.0*tmp_7) + 8.1380208333333333e-5*tmp_8*(4480.0*tmp_10 - 6656.0*tmp_4 - 14208.0) + 1.0


    cpdef double get(self, str force_name):
        # hard coded for now
        if force_name == "azimuthal":
            return self.FphiCorrMultParser
        elif force_name == "radial":
            return self.FrCorrMultParser
        raise RuntimeError(f"Unsupported force '{force_name}'")

    def initialize(self, *, flagPN1, flagPN2, flagPN3, flagPN32, flagPN52, nu):
        ret = self._initialize(flagPN1, flagPN2, flagPN3, flagPN32, flagPN52, nu)
        self._initialized = True
        return ret

    def compute(self, *, e, x, z):
        if not self._initialized:
            raise RuntimeError("Instance has not been initialized yet")
        return self._compute(e, x, z)